#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <time.h>
#include "coth.h"
#include "worker.h"
#include "sock.h"
#include "gfx.h"
#include "ihex.h"
#include "config.h"
#include "chip.h"
#include "sck.h"

  // these have a lot of margin. default timeout for the server
  // socket on my esp8266 is 180 seconds, which is 3 minutes.
  // JOB_TIME specifies the timeout for a response to a single
  // packet, not the whole job, so it's plenty.
#define IDLE_TIME (30*1000)
#define JOB_TIME  (5*1000)

/*****************************************************************
 The main thread:
  - manages the GUI
  - manages timers
  - does file I/O
  - sets up the connection

 The worker thread:
  - transfers data to/from the device according to the
    commands given by the user
  - closes the connection if there is an error
  - gives feedback to the user on the progress of work

 The worker thread will suspend its execution when it needs to
 read some data from the device. The main thread will do one
 of the following: 
  - let the worker thread continue with its execution when
    some data comes in.
  - tell the worker thread that the operation is cancelled
  - tell the worker thread that the operation has timed out

 In the latter two cases, the operation should be cancelled.
 In the last case, the socket should be closed and the user
 should make another connection to the device.
 *****************************************************************/

int worker_busy;
static int job_cancelled;
static int wo_func;
static void *wo_p1, *wo_p2;
static uint32_t wo_i1, wo_i2;

static uint32_t wo_idle_timer;
static uint32_t wo_job_timer;

static coth_t *comain, *coworker;

extern int device_socket;

static int  main_pipe;

static uint8_t param[1024];
static uint8_t send_buf[2048];
static uint8_t recv_buf[2048];
static uint8_t send_buf_len;
static uint32_t last_cmd_ident=1;

static int imin2(int a, int b) { return a<b? a:b; }

static int get_chip_msg(chip_t **R)
{
  *R= get_chip();
  if (!*R) { loprintf("please select a valid chip\n"); return 1; }
  return 0;
}


static int is_cancelled()
{
  if (job_cancelled) { job_cancelled= 0;return 1; }
  return 0;
}

static void signal_worker_completion(uint32_t i1,uint32_t i2,void *p1,void *p2)
{
  uint8_t c;
  main_i1= i1;
  main_i2= i2;
  main_p1= p1;
  main_p2= p2;
  write(main_pipe,&c,1);
}

void hex_print(uint16_t addr, uint8_t *data)
{
  int i;
  loprintf("%04x: ", addr);
  for(i=0;i<16;i++)
    loprintf("%02x ", data[i]);
  for(i=0;i<16;i++)
    if (data[i]>=' ' && data[i]<127) loprintf("%c",data[i]);
                                else loprintf(".");
  loprintf("\n");
}

void worker_run(int func, void *p1, void *p2, uint32_t i1, uint32_t i2)
{
  if (worker_busy)
    worker_cancel_job();
  wo_func= func;
  wo_p1= p1;
  wo_p2= p2;
  wo_i1= i1;
  wo_i2= i2;
  wo_idle_timer= 0;
  wo_job_timer= 0;
  coth_switch(comain, coworker);
}

static void update_firmware(char *fn, uint32_t boot_area_start);
static void save_eeprom(char *fn);
static void save_flash(char *fn);

static int command(int cmd, int plen)
{
  int i,R,L;
  uint32_t cmd_ident, id;
  send_buf_len= 0;

  b64_encode_begin(send_buf,&send_buf_len);
  for(i=0;i<32;i+=8) b64_encode(last_cmd_ident>>i);
  b64_encode(cmd);
  for(i=0;i<plen;i++) b64_encode(param[i]);
  b64_encode_end();

  send_buf[send_buf_len++]= '\r';
  send_buf[send_buf_len++]= '\n';

  cmd_ident= last_cmd_ident++;

  if (device_socket==-1 || is_cancelled()) return -1;
  if (sock_send(device_socket, send_buf, send_buf_len))
    { close_device(0); return -2; }

again:
  L= 0;
  while(1)
  {
    wo_job_timer= 0;
    coth_switch(coworker,comain);
    if (device_socket==-1 || is_cancelled())
      { gfx_logview_printf(Wlog, "command cancelled\n"); 
       return -2; }
    R= sock_recv(device_socket, recv_buf+L, sizeof(recv_buf)-L);
    if (R<0) { gfx_logview_printf(Wlog, "sock_recv returned %d\n",R);
               close_device(0); return -2; }
    for(i=0;i<R;i++) if (recv_buf[L+i]=='\r') break;
    L+= i;
    if (i!=R) break;
  }
  L= b64_decode(recv_buf, L);

  if (L<5) { gfx_logview_printf(Wlog, "Message too short\n"); return -2; }
  id= 0;
  for(i=0;i<4;i++) id= id | (recv_buf[i]<<(i*8));
  if (id<cmd_ident) goto again;
  if (id>cmd_ident) 
   { gfx_logview_printf(Wlog, "device is out of sync\n");
     return -2; }

  for(i=4;i<L;i++) param[i-4]= recv_buf[i];
  return L-4;
}

static void run_ident(int show)
{
  int R;
  R= command('I', 0);
  if (R<0) { gfx_logview_printf(Wlog, "Ident failed\n"); }
  else { param[R]= 0; if (show) 
          gfx_logview_printf(Wlog, "Ident: %s\n", param); }
}

static int param_encode_uint32(uint32_t V, int len)
{
  int i;
  for(i=0;i<4;i++) { param[len++]= V; V>>= 8; }
  return len;
}

static void set_baud_rate(uint32_t rate)
{
  int i,R;
  uint32_t V;
  i= param_encode_uint32(rate, 0);
  R= command('B', i);
  if (R<0 || param[0]!=0) gfx_logview_printf(Wlog, "set_baud_rate failed\n"); 
  else     gfx_logview_printf(Wlog,"set baud rate to %d",rate);
}

static void get_baud_rate()
{
  int R;
  R= command('b', 0);
  if (R<0) { gfx_logview_printf(Wlog, "get_baud_rate failed\n"); }
  else {
     uint32_t Z;
     Z= param[0] | (param[1]<<8) | (param[2]<<16) | (param[3]<<24);
     gfx_logview_printf(Wlog,"baud rate is: %d\n",Z);
     signal_worker_completion(Z, 0, NULL, NULL);
  }
}
        

static void change_ap_settings(char *ssid, char *pwd)
{
  int R;
  strcpy(param, ssid);
  R= command('S', strlen(ssid));
  if (R<0 || param[0]==0xff) 
    { gfx_logview_printf(Wlog, "Changing conf.ssid failed\n"); return ;}
  gfx_logview_printf(Wlog,"conf.ssid changed\n");

  strcpy(param, pwd);
  R= command('P', strlen(pwd));
  if (R<0 || param[0]==0xff) 
    { gfx_logview_printf(Wlog, "Changing conf.pwd failed\n"); return ;}
  gfx_logview_printf(Wlog,"conf.pwd changed\n");
}

#if 0
static void run_test(int addr)
{
  int i,j,R;
  uint8_t rdata[64];
  i= 0;
  param[i++]= addr;
  param[i++]= addr>>8;
  for(j=0;j<64;j++)
    param[i++]= rdata[j]= rand();
  gfx_logview_printf(Wlog, "Writing 64 random bytes " 
                           "starting at address %02x\n", addr);
  R= command('>', i);
  if (R<0 || param[0]==0xff)
   { gfx_logview_printf(Wlog,
      "Writing to external EEPROM failed (%s)\n", R<0? "R<0" : "0xff");
     return ;}
  for(j=0;j<i;j++) param[j]= 0;
  i= 0;
  param[i++]= addr;
  param[i++]= addr>>8;
  gfx_logview_printf(Wlog, "Reading back the written chunk\n");
  R= command('<', i);
  if (R<0 || param[0]==0xff)
   { gfx_logview_printf(Wlog, "Reading from external EEPROM failed (%s)\n"
                  , R<0? "R<0" : "0xff");
     return ;}
  for(j=0;j<64;j++)
   if (param[j+1]!=rdata[j])
   { gfx_logview_printf(Wlog, "Data read back doesn't match the sent data at"
                              "offset %d\n", j);
     return ; }
  gfx_logview_printf(Wlog, "All OK\n");
}
#endif

static int write_external_eeprom(uint16_t addr, uint8_t *buffer)
{
  int i,R;
  i= 0;
  param[i++]= addr;
  param[i++]= addr>>8;
  memcpy(param+i, buffer, 64);

  R= command('>', i+64);
  msleep(10); // 5 milliseconds is the maximum wait time between page writes
              // for 24xx256  
  if (R<0 || param[0]==0xff) return 1;
  return 0;
}

int verify_flash(uint16_t addr, uint8_t *data)
{
  int R;
  int i;
  i= 0;
  param[i++]= addr;
  param[i++]= addr>>8;
  R= command('?', i);

  if (R<0) return 1;

  for(i=0;i<64;i++)
   if (param[i]!=data[i])
   { gfx_logview_printf(Wlog, "Data read back doesn't match the sent data at"
                              "offset %d\n", i);
     return 1; }
  return 0;
}

static void port_op(char *pin, uint32_t op)
{
  int i,R;
  i= 0;
  param[i++]= pin[0];
  param[i++]= pin[1];
  param[i++]= op;
  param[i]= 0;
  gfx_logview_printf(Wlog,"port operation %s\n", param);
  R= command('/',3);
  if (R<0) gfx_logview_printf(Wlog,"failed\n");  
  else gfx_logview_printf(Wlog,"success\n");
}

static void run_test(int addr)
{
  int i,j,R;
  uint8_t rdata[128];

  for(i=0;i<128;i++) rdata[i]= rand();

  gfx_logview_printf(Wlog, "Writing 128 bytes to address %02x\n",addr);

  if (write_external_eeprom(addr, rdata))
     { gfx_logview_printf(Wlog, "First half failed\n"); return ;}

  if (write_external_eeprom(addr+64, rdata+64))
     { gfx_logview_printf(Wlog, "Second half failed\n"); return ;}

  gfx_logview_printf(Wlog,"'Booting' 1 page from external EEPROM\n");
  i= 0;
  param[i++]= addr;
  param[i++]= addr>>8;
  R= command('!', i);
  if (R<0 || param[0]==0xff)
     { gfx_logview_printf(Wlog, "Failed\n"); return ;}

  gfx_logview_printf(Wlog,"Verifying data in flash\n");
  i= 0;
  param[i++]= addr;
  param[i++]= addr>>8;

  if (verify_flash(addr, rdata))
     { gfx_logview_printf(Wlog, "Failed at first half\n"); return ;}

  if (verify_flash(addr+64, rdata+64))
     { gfx_logview_printf(Wlog, "Failed at second half\n"); return ;}


  gfx_logview_printf(Wlog, "All OK\n");
}

static int passthru_command(char *p,int sigcomplete)
{
  int R;
  strcpy(param, p+1);
  R= command(p[0], strlen(p+1));
  if (R<0) { loprintf("command %s failed\n",p); R= 1; }
      else R= 0;
  if (sigcomplete) signal_worker_completion(R,0,NULL,NULL);
  return R;
}

static int simple_command(char *msg, char *cmd)
{
  loprintf("%s\n", msg);
  return passthru_command(cmd,0);
}

static int pcommand(uint32_t cmd, int ret)
{
  int R;
  param[0]= cmd>>24;
  param[1]= cmd>>16;
  param[2]= cmd>>8;
  param[3]= cmd;
  param[4]= ret;
  R= command('C', 5);
  if (R<0) return -1;
  return param[0];
}

static int verify_signature()
{
  chip_t *C;
  int i,R;
  unsigned char sig[3];
  int j;

  C= get_chip();
  if (!C) { loprintf("please select a valid device first\n"); return 1; }  
  loprintf("Getting signature: ");
  for(i=0;i<3;i++)
  {
    R= pcommand(0x30000000 | (i<<8), 3);
    if (R==-1) { loprintf("FAILED\n"); return 1; }
    sig[i]= R;
  }
  for(i=0;i<3;i++) loprintf(" %02x", sig[i]); loprintf("\n");
  for(i=0;i<3;i++) if (C->sig[i]!=sig[i])  goto sig_mismatch;
  loprintf("Device %s verified\n", C->name);
  return 0;

sig_mismatch:
  loprintf("Signatures do not match\n");
  for(j=0;chips[j].name;j++)
  {
     for(i=0;i<3;i++) if (chips[j].sig[i]!=sig[i]) break;
     if (i==3)
     {
        loprintf("Installed device seems to be a %s\n", chips[j].name);
        break;
     }
  }
  loprintf("Selected device type was %s\n",C->name);
  return 1;
}

static int start_programming()
{
  int R;
  int sck;

  if (config.G_sck && config.sck>=0 && config.sck<=4)
      sck= config.sck;
  else
      sck= 3;

  loprintf("setting SCK frequency to %s\n", sck_values[sck].freq);
  param[0]= sck;
  R= command('a', 1);
  if (R<0) { loprintf("failed\n"); return 1; }

  if (simple_command("power off", "m0") ||  
      simple_command("select programmer power", "k1") ||
      simple_command("connect programming lines","j1") ||
      simple_command("pull reset", "r0")) return 1;
  msleep(100);
  if (simple_command("power on", "m1")) return 1;
  loprintf("begin programming: ");
  R= pcommand(0xac530000,2);
  if (R!=0x53) { loprintf("%02x FAIL\n", R&0xff); return 1; }
  loprintf("OK\n");

  if (verify_signature()) return 1;
  return 0;
}

static void end_programming()
{
  simple_command("power off", "m0");
  simple_command("select self power", "k0");
  simple_command("disconnect programming lines","j0");
  msleep(100);
  simple_command("power on", "m1");
}

void test_func(char *p1,char *p2, uint32_t i1, uint32_t i2)
{
  loprintf("start programming\n");
  start_programming();
  loprintf("end programming\n");
  end_programming();
}
static void program_device(char *flash_fn, char *eeprom_fn);

static void read_all_fuses();
static void write_all_fuses(uint32_t value,uint32_t old_value);

static void worker_main()
{
  while(1)
  {
    wo_idle_timer= 0;
    while (!wo_func) coth_switch(coworker, comain);
    worker_busy= 1;
    switch(wo_func)
    {
    case 'I': run_ident(wo_i1); break;
    case 'A': change_ap_settings(wo_p1, wo_p2); break;
    case 'T': update_firmware("../firmware/funky.hex", 0x7000);
              break;
    case 'F': update_firmware(wo_p1, wo_i1); break;
    case 'b': get_baud_rate(); break;
    case 'B': set_baud_rate(wo_i1); break;
    case '/': port_op(wo_p1, wo_i1); break;
    case 'G': passthru_command(wo_p1,wo_i1); break;
    case '?': test_func(wo_p1,wo_p2,wo_i1,wo_i2); break;
    case 'P': program_device(wo_p1,wo_p2); break;
    case 'u': read_all_fuses(); break;
    case 'U': write_all_fuses(wo_i1, wo_i2); break;
    case 'z': save_eeprom(wo_p1); break;
    case 'Z': save_flash(wo_p1); break;
    }
    wo_func= 0;
    worker_busy= 0;
  }
}

int worker_init()
{
  int fds[2];
  if (pipe(fds)) return -1;
  main_pipe= fds[1];
  comain= coth_main();
  coworker= coth_new(worker_main, 20*1024);
  job_cancelled= 0;
  wo_func= 0;
  worker_busy= 0;
  return fds[0];
}

void worker_cancel_job()
{
  if (!worker_busy) return ;
  job_cancelled= 1;
  coth_switch(comain, coworker);
}

void open_device(char *addr)
{
  if (device_socket!=-1) return ;
  device_socket= sock_connect(addr, 1011);
  if (device_socket<0) return ;
  gfx_monitor_file(device_socket, 30, 1, 0, 0);
}


void close_device(int canc)
{
  if (device_socket==-1) return ;
  gfx_stop_monitor_file(device_socket, 30);
  sock_close(device_socket);
  device_socket= -1;
  if (canc) worker_cancel_job();
}


void worker_read_data()
{
  coth_switch(comain, coworker);
}

void worker_time_elapsed(unsigned int ms)
{
  if (worker_busy)
    { wo_job_timer += ms;
      if (wo_job_timer>=JOB_TIME) {  gfx_logview_printf(Wlog,"cancelling job, it takes too long.\n"); worker_cancel_job(); } }
  else
    { wo_idle_timer += ms;
      if (wo_idle_timer>=IDLE_TIME && device_socket!=-1)
          worker_run('I', 0, 0, config.verbose_ident, 0); }
}


static void update_firmware(char *fn, uint32_t boot_area_start)
{
  char *e;
  uint8_t *flash_data;
  uint8_t pagemask[64];
  int i,j,R;
  if (boot_area_start<0x7000 || boot_area_start>0x7fff)
  {
    loprintf("Boot area start should be between 0x7000 and 0x7fff\n");
    return ;
  }
  e= read_hex_file(fn, 32*1024, 0xff, &flash_data);
  if (e)
  {
    loprintf("error reading hex file %s\n%s\n",fn,e);
    return ;
  }
  for(i=boot_area_start;i<0x7fff;i++)
    flash_data[i]= 0xff;

  memset(pagemask, 0, 64);
  gfx_logview_printf(Wlog, "Writing %s to external EEPROM\n",fn);
  gfx_progressbar_total(Wprogw, 32768);
  gfx_progressbar_total(Wprogr, 32768);
  gfx_progressbar_set(Wprogw, 0);
  gfx_progressbar_set(Wprogr, 0);
  for(i=0;i<32768;i+=128)
  {
    gfx_progressbar_set(Wprogw, i);
    for(j=0;j<128;j++)
      if (flash_data[i+j]!=0xff) break;
    if (j==128) continue; // empty page, skip it
    pagemask[ (i/128)/8] |= 1<<((i/128)%8);
    if (write_external_eeprom(i, flash_data+i) ||
        write_external_eeprom(i+64, flash_data+i+64))
    {
      gfx_logview_printf(Wlog, "External EEPROM write failed at page %d\n",
                     i/128);
      return ;
    }
    msleep(10); // for the EEPROM to get itself together
  }
  gfx_logview_printf(Wlog, "Writing page list to external EEPROM\n");
  if (write_external_eeprom(0x7f80, pagemask))
  {
      gfx_logview_printf(Wlog, "Page mask write to external EEPROM failed\n");
      return ;
  }
  gfx_progressbar_set(Wprogw, 32768);
  
  R= command('F', 0);
  if (R<0 || param[0]==0xff)
  {
      gfx_logview_printf(Wlog, "Boot command failed\n");
      return ;
  } 
  gfx_logview_printf(Wlog, "Disconnecting and rebooting now.\n"
                    "Please reconnect when booting done\n");

  close_device(0);
}

static int read_hex(int size,char *fn, uint8_t **Rd)
{
  char *e;
  *Rd= NULL;
  if (!fn) return 0;
  e= read_hex_file(fn,size,0xff,Rd);
  if (e) { loprintf("can't read input file %s:\n   %s\n",fn, e);
           return 1; }
  return 0;
}

static int program_flash_page(chip_t *C, uint16_t addr, uint8_t *data)
{
  int L,i,j,cp,R;
  uint16_t A;
  
  A= addr;
  for(L=0;L<C->flash_page_size;L+=cp)
  {
    A= addr+L;
    cp= C->flash_page_size - L;
    if (cp>64) cp= 64;
    j= 0;
    param[j++]= A;
    param[j++]= A>>8;
    param[j++]= cp;
    for(i=0;i<cp;i++) param[j++]= data[A+i];
    R= command('W', j);
    if (R<=0 || param[0]!=cp)
    {
       loprintf("error in loading flash page buffer at %04x\n",A);
       return 1;
    }
  } 

  R= pcommand( 0x4c000000 | (addr&~1)<<7, 4);

  if (R<0)
  {
    loprintf("error in writing page at %04x\n",addr);
     return 1;
  }
  msleep(10);
  return 0;
}

static int read_flash(chip_t *C, uint16_t A, uint8_t *data, int len)
{
  int i,j,R;
  j= 0;
  param[j++]= A;
  param[j++]= A>>8;
  param[j++]= len;

  R= command('R', j);
  if (R<=0 || param[0]!=len)  
  {
     loprintf("error in reading flash memory at %04x\n",A);
     return 1;
  }
  for(i=0;i<len;i++) data[i]= param[i+1];
  gfx_progressbar_set(Wprogr, A+len);
  return 0;
}

static int verify_flash_page(chip_t *C, uint16_t addr, uint8_t *data)
{
  int L,i,cp,R;
  uint8_t page[1024]; // should be enough
  
  for(L=0;L<C->flash_page_size;L+=cp)
  {
    cp= imin2(64,C->flash_page_size - L);
    if (read_flash(C, addr+L, page+L, cp)) return 1;
  } 
  for(i=0;i<C->flash_page_size;i++)
    if (page[i]!=data[addr+i])
    {
       loprintf("verification failed at address %02x\n",addr+i);
       int K= addr+i;
       K-= K%16;
       loprintf("written:\n");
       hex_print(K, data+K);
       loprintf("read:\n");
       hex_print(K, page+K);
       return 1;
    }

  return 0;
}

static int read_eeprom_all(chip_t *C, uint8_t **Rd)
{
  uint8_t *data;
  int size;
  int L,cp,i,R;
  size= C->eeprom_size;
  gfx_progressbar_total(Wprogr,size);
  gfx_progressbar_set(Wprogr,0);
  data= malloc(size);
  for(L=0;L<size;L+=cp)
  {
    cp= imin2(64, size-L);
    param[0]= L;
    param[1]= L>>8;
    param[2]= cp;
    R= command('e', 3);
    if (R<0 || param[0]!=cp) 
    {
       loprintf("EEPROM read failed at adress %04x\n",L); 
       free(data); 
       return 1; 
    }
    gfx_progressbar_set(Wprogr,L+cp);
    for(i=0;i<cp;i++) data[L+i]= param[i+1];
  }
  *Rd= data;
  return 0;
}

static int write_eeprom(chip_t *C, uint8_t *data, int start, int len)
{
  int i,j,cp,R;
  int A, end;

  end= start+len;
  for(A=start;A<end;A+=cp)
  {
    cp= imin2(64, end - A);
    j= 0;
    param[j++]= A;
    param[j++]= A>>8;
    param[j++]= cp;
    for(i=0;i<cp;i++) param[j++]= data[A+i];
    R= command('E', j);
    if (R<0 || param[0]!=cp)
    {
      loprintf("EEPROM write failed at address %04x\n",A);
      return 1;
    }
    gfx_progressbar_set(Wprogw, A+cp);
  }
  return 0;
}

static int m328_read_hifuse(uint8_t *RD)
{
  int R;
  R= pcommand(0x58080000, 3);
  if (R<0) { loprintf("read high fuse failed\n"); return -1; }
  *RD= R;
  loprintf("high fuse: %02x\n", R&0xff);
  return 0;
}


static int eeprom_preserved_across_erase(chip_t *C)
{
  if (chip_is(C,"Atmega328","Atmega328P","Atmega8A", "Atmega8", NULL))
  {
    uint8_t hifuse;
    loprintf("Reading high fuse byte to check EESAVE flag\n");
    if (m328_read_hifuse(&hifuse)) return 1;
    if (hifuse&(1<<3)) return 0;
    return 1;
  }
  return 1;
}

static int program_eeprom_read_first(chip_t *C, uint8_t *data)
{
  uint8_t *ddata;
  int s,e;
  loprintf("Reading EEPROM\n");
  if (read_eeprom_all(C, &ddata)) return 1;

  s= 0;
  gfx_progressbar_total(Wprogw,C->eeprom_size);
  gfx_progressbar_set(Wprogw,0);

  loprintf("Writing EEPROM\n");

again:
  for(;s<C->eeprom_size;s++)
    if (data[s]!=ddata[s]) break;

  gfx_progressbar_set(Wprogw, s);

  if (s==C->eeprom_size) goto success;

  for(e=s+1;e<C->eeprom_size;e++)
    if (data[e]==ddata[e]) break;

  if (write_eeprom(C, data, s,e-s)) goto fail;
  s= e;
  goto again;  

success: free(ddata); return 0;
fail: free(ddata); return 1;
}

static int program_eeprom_normal(chip_t *C, uint8_t *data)
{
  int s,e;

  s= 0;
  gfx_progressbar_total(Wprogw,C->eeprom_size);
  gfx_progressbar_set(Wprogw,0);

  loprintf("Writing EEPROM\n");

again:
  for(;s<C->eeprom_size;s++) if (data[s]!=0xff) break;

  gfx_progressbar_set(Wprogw, s);
  if (s==C->eeprom_size) return 0;

  for(e=s+1;e<C->eeprom_size;e++) if (data[e]==0xff) break;

  if (write_eeprom(C, data, s,e-s)) return 1;
  s= e;
  goto again;  
}

static void program_eeprom(chip_t *C, int chip_erased, uint8_t *data)
{
  uint8_t *ddata;
  int i;
  if (!chip_erased || eeprom_preserved_across_erase(C))
   { if (program_eeprom_read_first(C, data)) return ; }
  else
   { if (program_eeprom_normal(C, data)) return ; }
  loprintf("Verifying EEPROM\n");
  if (read_eeprom_all(C, &ddata)) return ;
  for(i=0;i<C->eeprom_size;i++)
    if (ddata[i]!=data[i])
    {
       loprintf("verification failed at address %02x\n",i);
       loprintf("written:\n");
       i-= i%16;
       hex_print(i, data+i);
       loprintf("read:\n");
       hex_print(i, ddata+i);
       break;
    }
  free(ddata);
}

static void program_flash(chip_t *C, uint8_t *data)
{
  int R,i,j;
  loprintf("Chip erase\n");   
  R= pcommand(0xac800000, 4);
  if (R<0) goto fail;
  msleep(5);
  
  gfx_progressbar_total(Wprogw,C->flash_size);
  gfx_progressbar_set(Wprogw,0);
  gfx_progressbar_total(Wprogr,C->flash_size);
  gfx_progressbar_set(Wprogr,0);

  loprintf("Writing flash\n");

  for(i=0;i<C->flash_size;i+=C->flash_page_size)
  {
    for(j=0;j<C->flash_page_size;j++)
      if (data[i+j]!=0xff) break;
    if (j!=C->flash_page_size)
      if (program_flash_page(C, i, data)) goto fail;
    gfx_progressbar_set(Wprogw, i+C->flash_page_size);
  }

  loprintf("Verifying flash\n");

  for(i=0;i<C->flash_size;i+=C->flash_page_size)
  {
    for(j=0;j<C->flash_page_size;j++)
      if (data[i+j]!=0xff) break;
    if (j!=C->flash_page_size)
      if (verify_flash_page(C, i, data)) goto fail;
  }

  return ;

fail:
  loprintf("failed\n");
}



static void program_device(char *flash_fn, char *eeprom_fn)
{
  chip_t *C;
  uint8_t *f_data, *e_data;
  char buf[200];
  struct tm now;
  time_t now_time;

  if (flash_fn) loprintf("program flash: %s\n", flash_fn);
  if (eeprom_fn) loprintf("program eeprom: %s\n", eeprom_fn);

  now_time= time(NULL);
  localtime_r(&now_time, &now);
  strftime(buf, sizeof(buf)-1, "%H:%M:%S", &now);
  loprintf("programming started at %s\n",buf);

  C= get_chip();
  if (!C) { loprintf("please select a valid device\n"); return ; }

  f_data= e_data= NULL;
  if (read_hex(C->flash_size,flash_fn, &f_data)) goto done;
  if (read_hex(C->eeprom_size,eeprom_fn, &e_data)) goto done;

  if (start_programming()) goto finish;

  if (f_data) program_flash(C, f_data);
  if (e_data) program_eeprom(C, f_data?1:0, e_data);

finish:
  end_programming();
done:
  if (f_data) free(f_data);
  if (e_data) free(e_data);
}

static int run_simple_pcommand
  (char *msg, uint32_t cmd, int idx, uint8_t *result)
{
  int R;
  loprintf("%s: ", msg);
  R= pcommand(cmd,idx);
  if (R<0) { loprintf("FAIL\n"); return 1; }
  loprintf("%02x\n", R&0xff);
  *result= R;
  return 0;
}

static void read_all_fuses()
{
  chip_t *C;
  int FV;
  uint8_t fuses[3];
  FV= -1;

  fuses[0]= fuses[1]= fuses[2]= 0;
  C= get_chip();
  if (!C) { loprintf("no device was selected\n"); goto done; }
  if (start_programming()) goto finish;
  if (chip_is(C,"Atmega328P","Atmega328",NULL))
  {
    if (run_simple_pcommand("read low fuse", 0x50000000, 3, fuses+0) ||
        run_simple_pcommand("read high fuse", 0x58080000, 3, fuses+1) ||
        run_simple_pcommand("read extended fuse",0x50080000, 3, fuses+2))
     goto finish;
  }
  else if (chip_is(C,"Atmega8A","Atmega8",NULL))
  {
    if (run_simple_pcommand("read low fuse", 0x50000000, 3, fuses+0) ||
        run_simple_pcommand("read high fuse", 0x58080000, 3, fuses+1))
     goto finish;
  }
  FV= (fuses[2]<<16) | (fuses[1]<<8) | fuses[0];

finish:
  end_programming();  
done: 
  signal_worker_completion(FV, 0, NULL, NULL);
}

static void write_all_fuses(uint32_t value,uint32_t old_value)
{
  uint8_t fuses[3];
  uint8_t ofuses[3];
  chip_t *C;
  uint32_t cmdseq_m328[3]= { 0xACA00000, 0xACA80000, 0xACA40000 };
  uint32_t cmdseq_m8[2]=   { 0xACA00000, 0xACA80000 };
  int i,R,len;
  uint32_t *seq;

  C= get_chip();
  if (!C) { loprintf("no device was selected\n"); return ; }

  fuses[0]= value; fuses[1]= value>>8; fuses[2]= value>>16;
  ofuses[0]= old_value; ofuses[1]= old_value>>8; ofuses[2]= old_value>>16;

  for(i=0;i<3;i++) if (fuses[i]!=ofuses[i]) break;
  if (i==3) { loprintf("write fuse: nothing to do\n"); return ; }

  if (chip_is(C,"Atmega328P","Atmega328",NULL))
    { seq= cmdseq_m328; len= 3; }
  else if (chip_is(C,"Atmega8A", "Atmega8",NULL))
    { seq= cmdseq_m8; len= 2; }

  static char *names[3]= { "low", "high", "extended" };
  if (start_programming()) goto finish;
  for(i=0;i<len;i++)
  {
   if (fuses[i]==ofuses[i]) continue;
   loprintf("write %s fuse byte %02x: ", names[i], fuses[i]);
   R= pcommand(seq[i] | fuses[i], 4);
   if (R<0) loprintf("FAIL\n");
      else loprintf("OK\n");
  }
   
finish:
  end_programming();
}

static void write_hex(char *fn, uint8_t *data, int size)
{
  char *e;
  if (!data) { loprintf("empty data, internal error\n"); return ; }
  loprintf("writing file %s\n", fn);
  e= write_hex_file(fn, size, 0xff, data);
  if (e) { loprintf("error writing file: %s\n", e); }
     else { loprintf("Write OK\n"); }
  
done:
  if (data) free(data);
}

static void save_eeprom(char *fn)
{
  chip_t *C;
  uint8_t *data;
  int ok;
  data= NULL;
  if (get_chip_msg(&C)) return ;

  loprintf("Saving EEPROM to %s\n",fn);
  ok= 0;
  if (start_programming()) goto finish;
  if (read_eeprom_all(C, &data)) goto finish;
  ok= 1;
  
finish:
  end_programming();
  if (ok) write_hex(fn, data, C->eeprom_size); 
}

static void save_flash(char *fn)
{
  chip_t *C;
  uint8_t *data;
  int ok,A;
  data= NULL;
  if (get_chip_msg(&C)) return ;

  ok= 0;
  loprintf("Saving flash to %s\n",fn);
  if (start_programming()) goto finish;
  gfx_progressbar_total(Wprogr,C->flash_size);
  gfx_progressbar_set(Wprogr,0);
  data= malloc(C->flash_size);
  for(A=0;A<C->flash_size;A+=64)
    if (read_flash(C, A, data+A, 64)) goto finish;
  ok= 1;
  
finish:
  end_programming();
  if (ok) write_hex(fn, data, C->flash_size); 
}


