
/***********************************************************************
The below module implements the i2c slave functionality thru interrupts.
You should call i2c_slave_init() when ready for I2C operation.
Address of the device is defined in I2C_ADDRESS macro.
When finished with i2c, i2c_slave_finish() should be called.

During transfers, the following functions will be called:

int i2c_slave_set_pointer(unsigned char A);

 A is the address of the element to be read/written to in the next
 operation. This function is called as part of a write transaction
 to slave. If your program isn't able to receive bytes at address A,
 the return value should be 0. Otherwise it should be 1.

int i2c_slave_received_data(unsigned char D);

 One byte of data has been received from the master. The pointer
 should be incremented. The return value is 1 if your application
 can receive more data, 0 otherwise.

int i2c_slave_start_send(unsigned char *D);

 This function is called when a read transaction starts. The data
 should be returned in *D. The return value should be 1 if there
 is more data to be sent, 0 otherwise.

int i2c_slave_send(unsigned char *D);

 This function is called for subsequent bytes to be sent to master.
 The return value is the same as start_send().

Both start_send() and send() should increment the pointer.

 ***********************************************************************/

#include <util/twi.h>


static unsigned char i2cslave_have_pointer;

void i2c_slave_init()
{
  TWAR= I2C_ADDRESS << 1;
  TWCR = (1<<TWEA) | (1<<TWEN) | (1<<TWIE) | (1<<TWINT);
}

/********************

 status register:
 7 6 5 4 . 3 2 1 0
 8 4 2 1 . 8 4 2 1 
 ---------------
 S S S S . S - P P    = f8


********************/


ISR(TWI_vect)
{
  unsigned char D;
  int R;
  switch(TW_STATUS)
  {
  // OWN SLA+W has been received. ACK has been returned.
  // we get the next byte for address, and return an ack.
  case TW_SR_SLA_ACK: 
    i2cslave_have_pointer= 0;
    TWCR = (1<<TWINT) | (1<<TWEA) | (1<<TWEN) | (1<<TWIE);
    break;
  case TW_SR_DATA_ACK:
  // This happens after we do the actions in 0x60. 
  // Bytes will be received in this status code.
    D= TWDR;
    if (i2cslave_have_pointer) R= i2c_slave_received_data(D); 
                 else { R= i2c_slave_set_pointer(D); i2cslave_have_pointer=1; }
    if (R) TWCR = (1<<TWINT) | (1<<TWEA) | (1<<TWEN) | (1<<TWIE);
      else TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWIE);
    break;
  case TW_SR_DATA_NACK:
  // in the case above  (0x80), we told TWI to send a NACK to the
  // next byte received. Now we have received that byte and sent
  // a NACK. All we need to do is to process this byte and then
  // return to an unaddressed state.
    D= TWDR;
    if (i2cslave_have_pointer) i2c_slave_received_data(D); 
                 else { i2c_slave_set_pointer(D); i2cslave_have_pointer=1; }
    TWCR = (1<<TWINT) | (1<<TWEA) | (1<<TWEN) | (1<<TWIE);
    break;
  case TW_SR_STOP:
  // A stop condition or a repeated start condition has been
  // detected. A repeated start happens when the master wants
  // to read something. This works as follows. The master first
  // sends start+device_address+W. That is followed by a memory
  // address. This address is stored by the slave. After
  // this, a repeated start+device_address+R is sent by the
  // master. Now, the slave can send data. So, here we simply
  // switch to transmitter mode. 
    TWCR = (1<<TWINT) | (1<<TWEA) | (1<<TWEN) | (1<<TWIE);
    break;
  // Rest of the receiver status values relate to multi-master operation.
  // I won't do that.
  // Now the transmitter stuff happens.

  // We received SLA+R and sent ACK. Now it's our turn to send data.
  case TW_ST_SLA_ACK:
     R= i2c_slave_start_send(&D);
     TWDR= D;
     if (R) TWCR = (1<<TWINT) | (1<<TWEA) | (1<<TWEN) | (1<<TWIE);
       else TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWIE);
     break;
   // Data has been sent, ACK was been returned.
   case TW_ST_DATA_ACK:
     R= i2c_slave_send(&D);
     TWDR= D; 
     if (R) TWCR = (1<<TWINT) | (1<<TWEA) | (1<<TWEN) | (1<<TWIE);
       else TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWIE);
     break;
   // Data was transmitted, a NACK was returned. We're ending the
   // transmission here.
   case TW_ST_DATA_NACK:
     TWCR= (1<<TWINT) | (1<<TWEA) | (1<<TWEN) | (1<<TWIE);
     break;
   // Now this is critical. We were expecting a NACK, but got an ACK
   // instead. ie. we ran out of things to send to the master, but
   // they still expect some bits from us. The data sheet says that
   // a bogus byte of all 1s will be sent in this case and there is
   // nothing I can do. Yet again, we simply switch to unaddressed
   // mode.
   case TW_ST_LAST_DATA:
     TWCR= (1<<TWINT) | (1<<TWEA) | (1<<TWEN) | (1<<TWIE);
     break;
   default:
     TWCR= (1<<TWINT) | (1<<TWIE) | (1<<TWEN);
   // case 0xF8:
   //   break;
   // all other status values are irrelevant. I will simply stop the
   // whole thing and start over.
   // default:
     // TWCR= (1<<TWINT) | (1<<TWEN) | (1<<TWSTO);
     // break;
  }
}

void i2c_slave_finish()
{
  TWCR &= ~(1<<TWEN);
}



