#include "config.h"
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include "def.h"
#include "usb.h"
#include "fiface.h"

#define PM328_VCC   (B,1)
#define PM328_MOSI  (D,5)
#define PM328_MISO  (D,6)
#define PM328_SCK   (D,7)
#define PM328_RESET (B,0)

#include "prg_mega328.c"

uint8_t buf[5];
uint8_t buf_len;

volatile uint8_t Dbuffer[32], Dhead, Dtail;
volatile uint8_t Pbuffer[32], Phead, Ptail;

uint8_t o_fn, i_fn, cmd_result;

uint16_t p_address;
uint16_t r_len;

void i2c_slave_init();
void i2c_slave_finish();

uint8_t Pavail()
{
  if (Ptail>=Phead) return 31-(Ptail-Phead);
  return Phead-Ptail-1;
}

uint8_t Plen()
{
  if (Ptail>=Phead) return Ptail-Phead;
  return 32-Phead+Ptail;
}

uint8_t Davail()
{
  if (Dtail>=Dhead) return 31-(Dtail-Dhead);
  return Dhead-Dtail-1;
}

uint8_t Dlen()
{
  if (Dtail>=Dhead) return Dtail-Dhead;
  return 32-Dhead+Dtail;
}

byte_t get_Pbytes(byte_t *data, byte_t len)
{
  byte_t i;
  for(i=0;i<len;i++)
  {
    if (Phead==Ptail) break;
    data[i]= Pbuffer[Phead];
    Phead= (Phead+1)%32;
  }
  return i;
}



   // return the data requested by the usb transfer.
uint8_t usb_in(byte_t* data, byte_t len)
{
  if (i_fn=='F')
  {
    if (len>r_len) len= r_len;
    Pm328_read_flash(p_address,data,len);
    p_address+= len;
    r_len -= len;
    return len;
  }
  if (i_fn=='T') return get_Pbytes(data, len);
  return 0;
}

void record_command_bytes(byte_t *data, byte_t len)
{
  int i;
  for(i=0;i<len && buf_len<5;i++)
  {
    buf[buf_len]= data[i];
    buf_len++;
  }
  if (buf_len==5)
  {
    cmd_result= Pm328_send_command(buf, buf[4]);
    o_fn= 0;
    buf_len= 0;
  }
}

void record_Dbytes(byte_t *data, byte_t len)
{
  int i;
  for(i=0;i<len;i++)
  {
    if ((Dtail+1)%32==Dhead) break;
    Dbuffer[Dtail]= data[i];
    Dtail= (Dtail+1)%32;
  }
}

   // use the data specified by the usb transfer
void usb_out(byte_t* data, byte_t len)
{
  switch(o_fn)
  {
  case 'C': record_command_bytes(data,len); break;
  case 'F': Pm328_load_flash_buffer(p_address,data, len);
            p_address+= len;
            break;
  case 'T': record_Dbytes(data, len); break;
  }
}

static const char test_str[]= "USBp328";

byte_t usb_setup(byte_t data[8])
{
  uint8_t req;
  uint8_t i;
  // uint16_t value;
  // the setup packet is structured as follows:
  // offset size meaning
  //  0      1   request type, describes the endpoint direction etc.
  //  1      1   request code. any value is OK, since the above
  //                           field already specifies VENDOR and
  //                           we're free to use any value.
  //  2      2   value    these two fields are generic fields for
  //  4      2   index    identifying stuff. I just use them as data fields
  //  6      2   length of data to follow in the _out() packets
  req= data[1];
  // if there is data to be returned, we can do that within the data[]
  // array, using up to 8 bytes. If there are more, we should return
  // 0xff and usbtiny will call usb_in().
  switch(req)
  {
case CMD_TEST:               for(i=0;i<7;i++) data[i]= test_str[i]; return i;
case CMD_ENTER_PRGMODE:      
  i2c_slave_finish();
  Pm328_enter_programming_mode();
  return 0;
case CMD_LEAVE_PRGMODE:
  Pm328_leave_programming_mode();
  i2c_slave_init();
  return 0;
case CMD_PRG_COMMAND:        o_fn= 'C'; buf_len= 0; return 0;
case CMD_PRG_COMMAND_RESULT: data[0]= cmd_result; return 1;
case CMD_SET_ADDRESS:        p_address= data[2] | (data[3]<<8); return 0;
case CMD_WRITE_PAGEBUFFER:   o_fn= 'F'; return 0;
case CMD_READ_FLASH:         r_len= data[2] | (data[3]<<8); i_fn= 'F'; 
                             return 0xff;
case CMD_RESET_TARGET:       Pm328_recycle_vcc(); return 0;
case CMD_GET_WSIZE:          data[0]= Davail(); return 1;
case CMD_GET_RSIZE:          data[0]= Plen(); return 1;
case CMD_DIO_SEND:           o_fn= 'T'; return 0;
case CMD_DIO_RECV:           i_fn= 'T'; return 0xff;
case CMD_QUEUEPTRS:          data[0]= Dhead; data[1]= Dtail; 
                             data[2]= Phead; data[3]= Ptail;
                             return 4;
case CMD_SETSPEED:           Pm328_speed= data[2]; return 0;
case CMD_RESETPTRS:          Dhead= Dtail= Phead= Ptail= 0; return 0;
  }
  return 0; 
}

uint8_t dio_addr;

int i2c_slave_set_pointer(unsigned char A)
{
  dio_addr= A;
  return 1;
}

int i2c_slave_received_data(unsigned char D)
{
  if (dio_addr!=DIO_REG_BUFFER) return 0;
  if ((Ptail+1)%32==Phead) return 0;
  Pbuffer[Ptail]= D;
  Ptail= (Ptail+1)%32;
  if (Ptail==Phead) return 0;
  return 1;
}

#define i2c_slave_start_send   i2c_slave_send

int i2c_slave_send(unsigned char *D)
{
  if (dio_addr==DIO_REG_WSIZE)  { *D= Pavail(); return 1; }
  if (dio_addr==DIO_REG_RSIZE)  { *D= Dlen();   return 1; }
  if (dio_addr!=DIO_REG_BUFFER) { *D= 0xff;     return 0; }
  if (Dhead==Dtail) { *D= 0xff; return 0; } 
  *D= Dbuffer[Dhead];
  Dhead= (Dhead+1)%32;
  if (Dhead==Dtail) return 0;
  return 1;
}

#define I2C_ADDRESS PCOM_ADDR
#include "i2cslave.c"

int main()
{
   usb_init();
   i2c_slave_init();
   while(1) { usb_poll(); }
}

