#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <readline/readline.h>
#include <readline/history.h>

#include "../firmware/fiface.h"
#include "iface.h"
#include "cmdline.h"
#include "ihex.h"
#include "hexdump.c"

static int verbose;

char **get_line()
{
  char cmd[1000];
  char **args;
again:
  fgets(cmd, sizeof(cmd)-1, stdin);
  cmd[sizeof(cmd)-1]= 0;
  if (!cmd[0]) { strcpy(cmd, "quit\n"); }

  args= parse_cmd_line(cmd);
  if (!args) goto again;
  return args;
}

void put_file(char *fn)
{
  char *e;
  int npages;
  int i;
  device_page_t *pages;
  uint16_t A;
     /*******************************************************
                     W A R N I N G
      The intel hex file format uses byte addresses and
      byte counts. The resulting addresses in returned
      pages should be divided by two to get word addresses.

      Also, the page size here is a byte count. Since we
      operate on a byte-by-byte basis in the firmware, this
      is fine.
                     W A R N I N G
      *******************************************************/
  e= read_hex_file(fn, 128, 32767, 0xff, &npages, &pages);
  if (e)
  {
    printf("error in file %s\n", fn);
    printf("%s\n",e);
    return ;
  }
  uint8_t buf[5];
  for(i=0;i<npages;i++)
  {
    A= pages[i].address;
    printf("load page %04x\n", A);
    Usimple1(CMD_SET_ADDRESS, A);
     int j;
    for(j=0;j<128;j++)
    {
      if (j%16==0) { if (j) printf("\n");
                     printf("%04x>", A+j); }
      printf("%02x", pages[i].data[j]);
    }
    printf("\n");
    for(j=0;j<128;j+=16)
      Uwrite0(CMD_WRITE_PAGEBUFFER, pages[i].data+j, 16);
    printf("write page %04x\n", A);
    buf[0]= 0x4c;
    buf[1]= A>>9;
    buf[2]= A>>1;
    buf[3]= 0x00;
    buf[4]= 3;
    Uwrite0(CMD_PRG_COMMAND, buf, 5);
    msleep(50);
  }
  printf("all done\n");
  for(i=0;i<npages;i++) free(pages[i].data);
  free(pages);
}

uint8_t pcmd(uint32_t C, int idx)
{
  uint8_t buf[5];
  buf[0]= C>>24; buf[1]= C>>16; buf[2]= C>>8; buf[3]= C;
  buf[4]= idx;
  Uwrite0(CMD_PRG_COMMAND, buf, 5);
  if (idx==4) return 0;
  msleep(1);
  Uread0(CMD_PRG_COMMAND_RESULT, buf, 1);
  return buf[0];
}

typedef struct 
{
  uint8_t data[3];
  char *name;
} signature_t;

signature_t signatures[]=
 {
    { { 0x1e, 0x95, 0x14 }, "Atmega328" },
    { { 0x1e, 0x95, 0x0f }, "Atmega328P" },
    { { 0x00, 0x00, 0x00 },  NULL }
 };

void print_bufline(char *prompt, uint16_t addr, uint8_t *data)
{
  int i;
  printf("%s> %04x: ", prompt, addr);
  for(i=0;i<16;i++) printf("%02x",data[i]);
  printf("\n");
}

void progress(int amount, int total)
{
  int nc,i;
  if (verbose) return ;
  if (total==0) return ;
  nc= (amount*64)/total;
  printf("["); for(i=0;i<nc;i++) printf("#");
  for(;i<64;i++) printf(" ");
  printf("]\r");
  fflush(stdout);
}
 

void program_file(char *fn,char *efn)
{
  uint8_t R;
  int i,j,k,cp;
  uint8_t sig[3];
  char *e;
  int npages,neepages;
  device_page_t *pages,*eepages;
  uint16_t A;
  uint8_t buf[128];

  pages= NULL;
  eepages= NULL;
  neepages= 0;

  printf("Enter programming mode\n");
  Usimple0(CMD_ENTER_PRGMODE);

  printf("Programming Enable: ");
  fflush(stdout);
  R= pcmd(0xac530000, 2);
  printf("%02x ", R);
  if (R==0x53) printf("OK\n");
        else { printf("FAIL\n"); goto leave_pmode; }

  printf("Signature: "); fflush(stdout);
  for(i=0;i<3;i++)
  {
    sig[i]= pcmd( 0x30000000 | (i<<8), 3 );
    printf(" %02x", sig[i]); fflush(stdout);
  }

  for(i=0;signatures[i].name;i++)
  {
    for(j=0;j<3;j++) if (signatures[i].data[j]!=sig[j]) break;
    if (j==3)
    {
      printf(" = %s\n", signatures[i].name);
      break;
    }
  }
  if (!signatures[i].name)
  {
     printf("NOT FOUND\n");
     goto leave_pmode;
  }

     /*******************************************************
                     W A R N I N G
      The intel hex file format uses byte addresses and
      byte counts. The resulting addresses in returned
      pages should be divided by two to get word addresses.
                     W A R N I N G
      *******************************************************/
  e= read_hex_file(fn, 128, 32767, 0xff, &npages, &pages);
  if (e)
  {
    printf("error in file %s\n", fn);
    printf("%s\n",e);
    goto leave_pmode;
  }

  if (efn)
  {
    e= read_hex_file(efn, 1024, 1023, 0xff, &neepages, &eepages);
    if (e)
    {
      printf("error in file %s\n", fn);
      printf("%s\n",e);
      goto leave_pmode;
    }
  }

  printf("Chip erase\n");   pcmd(0xac800000, 4); msleep(5);

  printf("Writing %d pages\n", npages);
  int total;
  total= npages*128;
  progress(0,total);
  for(i=0;i<npages;i++)
  {
    A= pages[i].address;
    if (verbose) printf("Load page  %04x\n", A);
    Usimple1(CMD_SET_ADDRESS, A);
    for(j=0;j<128;j+=16)
    {
      Uwrite0(CMD_WRITE_PAGEBUFFER, pages[i].data+j, 16);
      progress(i*128+j, total);
    }

    if (verbose) printf("write page %04x\n", A);
    pcmd( 0x4c000000 | (A&~1)<<7, 4);
    msleep(10);
  }
  progress(total,total);
  if (!verbose) printf("\n");
  printf("Write complete\n");

  printf("Verifying %d pages\n", npages);
  progress(0,total);
  for(i=0;i<npages;i++)
  {
    A= pages[i].address;
    if (verbose) printf("Verifying page %04x\n", A);
    Usimple1(CMD_SET_ADDRESS, A);
    for(j=0;j<128;)
    {
      cp= 128-j;
      if (cp>16) cp= 16;
      k= Uread1(CMD_READ_FLASH,16, buf+j, cp);
      if (k>0) { j+= k; progress(i*128+j,total); }
          else { printf("Read flash failed\n"); goto leave_pmode; }
    }
    for(j=0;j<128;j++)
      if (pages[i].data[j]!=buf[j])
      {
        printf("verification failed at address %04x\n", A+j);
        j%= 16;
        print_bufline("write", A+j, pages[i].data+j);
        print_bufline(" read", A+j, buf+j);
        goto leave_pmode;
      }
  }
  if (!verbose) printf("\n");

  if (!neepages) goto skip_eeprom;

  int count= 0, t;
  for(j=0;j<1024;j++) if (eepages[0].data[j]!=0xff) count++;
  if (count==0) { printf("empty eeprom file, no need to program it\n");
                  goto skip_eeprom; }
  
  printf("Writing %d bytes to EEPROM\n",count);
  progress(0,count);
  t= 0;
  for(j=0;j<1024;j++)
    if (eepages[0].data[j]!=0xff)
    {
      pcmd(0xc0000000 | (j<<8) | eepages[0].data[j], 4);
      t++;
      progress(t, count);
      msleep(10); // normally, only 3.6
    }
  printf("\n");

  printf("Verifying %d bytes from EEPROM\n",count);
  progress(0,count);
  t= 0;
  for(j=0;j<1024;j++)
    if (eepages[0].data[j]!=0xff)
    {
      uint8_t Z;
      Z= pcmd(0xa0000000 | (j<<8) , 3);
      t++;
      progress(t, count);
      if (Z!=eepages[0].data[j])
      {
        printf("\nEEPROM verify failed at address %02x.\n"
               "Wrote %02x, got back %02x.\n", j, eepages[0].data[j], Z);
        goto leave_pmode; 
      } 
    }
  printf("\n");

skip_eeprom:
  printf("All OK\n");

leave_pmode:
  printf("Leave programming mode\n");
  Usimple0(CMD_LEAVE_PRGMODE);

  if (pages)
  {
    for(i=0;i<npages;i++) free(pages[i].data);
    free(pages);
  }
}

void read_eeprom()
{
  int i,j;
  uint8_t buf[16];
  printf("Reading 1024 bytes from EEPROM, skipping all 0xff lines.\n");
  for(j=0;j<1024;j+=16)
  {
    for(i=0;i<16;i++)
      buf[i]= pcmd(0xa0000000  | ((i+j)<<8), 3);
    for(i=0;i<16;i++) if (buf[i]!=0xff) break;
    if (i!=16) hexprint(buf, 16, j, 4);
  }
  printf("FINISHED\n");
}

void read_flash()
{
  int i,j,prt;
  uint8_t buf[40000];
  Usimple1(CMD_SET_ADDRESS, 0);

  prt= 0;
  for(i=0;i<32768;)
  {
    int k;
    k= Uread1(CMD_READ_FLASH,16, buf+i, 16);
    if (k>0) i+= k;
    while(i-prt>=16)
    {
      for(j=0;j<16;j++) if (buf[prt+j]!=0xff) break; 
      if (j!=16) {
      printf("%04x: ", prt);
      for(j=0;j<16;j++) printf("%02x ",buf[prt+j]);
      printf(" ");
      for(j=0;j<16;j++) 
        if (isprint(buf[prt+j])) printf("%c",buf[prt+j]);
                            else printf(".");
      printf("\n");
      }
      prt+= 16;
    }
  }
}

int have_char_at_terminal()
{
  fd_set rset;
  int r;
  struct timeval tm;
  FD_ZERO(&rset);
  FD_SET(0, &rset);
  tm.tv_sec= 0;
  tm.tv_usec= 50000;
  r= select(1, &rset, NULL, NULL, &tm);
  if (r>0) return 1;
  return 0;
}

int main()
{
  char **args;
  char *file_name,*eefile_name;

  eefile_name= file_name= NULL;
  Uinit(3);

  while(1)
  {
    args= get_line();
    if (!strcmp(args[0], "pp"))
    {
      program_file(file_name,eefile_name);
      continue;
    }
    if (!strcmp(args[0], "eras"))
   {
     uint8_t buf[5]; buf[0]= 0xac; buf[1]= 0x80; buf[2]= 0x00; buf[3]= 0x00;
     buf[4]= 0x00;
     Uwrite0(CMD_PRG_COMMAND, buf, 5);
     continue;
    }
    if (!strcmp(args[0],"test"))
    {
      char buf[20];
      memset(buf, 0, sizeof(buf));
      Uread0(CMD_TEST, (uint8_t*) buf, 20);
      printf("result: %s\n", buf);
      continue;
    }
    if (!strcmp(args[0],"quit")) goto done;
    if (!strcmp(args[0],"file"))
    {
      if (!args[1]) {
         printf("ERROR: usage: file <file_name> [<eeprom_file_name>]\n");
         continue;
      }
      file_name= strdup(args[1]);
      if (file_name[0] && args[2]) eefile_name= strdup(args[2]);
              else eefile_name= NULL;
      if (file_name) printf("file: <%s>\n", file_name);
               else printf("no file\n");
      if (eefile_name) printf("EEPROM file: <%s>\n", eefile_name);
               else printf("no EEPROM file\n");
      continue;
    }
    if (!strcmp(args[0],"ep"))
    {
      printf("ep: %d\n",Usimple0(CMD_ENTER_PRGMODE));
      continue;
    }
    if (!strcmp(args[0],"bp"))
    {
      uint8_t buf[5];
      printf("sending 'begin programming\n");
      buf[0]= 0xac; buf[1]= 0x53; buf[2]= 0x00; buf[3]= 0x00; buf[4]= 2;
      Uwrite0(CMD_PRG_COMMAND, buf, 5);
      msleep(5);
      Uread0(CMD_PRG_COMMAND_RESULT, buf, 1);
      printf("result: %02x\n", buf[0]);
      continue;
    }
    if (!strcmp(args[0],"lp"))
    {
      printf("lp: %d\n",Usimple0(CMD_LEAVE_PRGMODE));
      continue;
    }
    if (!strcmp(args[0],"sig"))
    {
      int i;
      uint8_t buf[5];
      uint8_t R[1];
      buf[0]= 0x30; buf[1]= 0x00; buf[3]= 0x00; buf[4]= 3;
      for(i=0;i<3;i++)
      {
        buf[2]= i;
        Uwrite0(CMD_PRG_COMMAND, buf, 5);
        msleep(5);
        Uread0(CMD_PRG_COMMAND_RESULT, R, 1);
        printf("sig: %02x\n", R[0]);
      }
      continue;
    }
    if (!strcmp(args[0],"prg"))
    {
      if (!file_name) { printf("no file\n"); continue; }
      put_file(file_name);
      continue;
    }
    if (!strcmp(args[0],"read"))
    {
      read_flash();
      continue;
    }
    if (!strcmp(args[0],"reset"))
    {
      Usimple0(CMD_RESET_TARGET);
      continue;
    }
    if (!strcmp(args[0],"ds"))
    {
      Uwrite0(CMD_DIO_SEND, (uint8_t*) args[1], strlen(args[1]));
      continue;
    }
    if (!strcmp(args[0],"dr"))
    {
      uint8_t buf[32];
      int r;
      r= Uread0(CMD_DIO_RECV, buf, 16);
      printf("recv(%d) ",r);
      if (r<=0) { printf("\n"); continue; }
      buf[r]= 0;
      printf("<%s>\n", buf);
      continue;
    }
    if (!strcmp(args[0],"tc"))
    {
      int lanl,i;
      printf("--test command <%s>\n",args[1]);
      Uwrite0(CMD_DIO_SEND, (uint8_t*) args[1], strlen(args[1]));
      lanl= 0;
      while(1)
      {
         uint8_t buf[32];
         int r;
         if (have_char_at_terminal()) { printf("\n--stopped\n"); break; }
         r= Uread0(CMD_DIO_RECV, buf, 16);
         for(i=0;i<r;i++)
         { if (buf[i]=='\n') { if (lanl) {lanl= 2; break;}
                                 lanl= 1; printf("\\n"); }
                        else lanl= 0; 
           if (buf[i]=='\r') printf("\\r");
           else printf("%c", buf[i]);
         }
         if (lanl==2) break;
      }
      printf("--end\n");
      continue;
    }
    if (!strcmp(args[0],"qp"))
    {
      uint8_t buf[32];
      int r;
      r= Uread0(CMD_QUEUEPTRS, buf, 16);
      printf("recv(%d) Dhead= %d Dtail= %d Phead= %d Ptail= %d\n",r, buf[0], buf[1], buf[2], buf[3]);
      continue;
    }
    if (!strcmp(args[0],"speed"))
    {
      Usimple1(CMD_SETSPEED, atoi(args[1]));
      continue;
    }
    if (!strcmp(args[0],"verbose"))
    {
      if (args[1]) { verbose= atoi(args[1]); } else verbose= 1;
      continue;
    }
    if (!strcmp(args[0], "rfuse"))
    {
      printf("read fuse: %02x\n", pcmd(0x50000000, 3));
      continue; 
    }
    if (!strcmp(args[0], "rfuseh"))
    {
      printf("read fuse high: %02x\n", pcmd(0x58080000, 3));
      continue; 
    }
    if (!strcmp(args[0], "rfusex"))
    {
      printf("read fuse extended: %02x\n", pcmd(0x50080000, 3));
      continue; 
    }
    if (!strcmp(args[0], "wfuse"))
    {
      uint8_t fb;
      char k[200];
      fb= strtol(args[1], NULL, 16);
      printf("write fuse low byte with value 0x%02x? (Y/N): ",fb);
      fflush(stdout);
      memset(k,0,sizeof(k));
      fgets(k,sizeof(k),stdin);
      if (k[0]=='y' || k[0]=='Y')
      {
        pcmd(0xACA00000| fb,4);
        continue;
      }
    }
    if (!strcmp(args[0], "wfuseh"))
    {
      uint8_t fb;
      char k[200];
      fb= strtol(args[1], NULL, 16);
      printf("write fuse high byte with value 0x%02x? (Y/N): ",fb);
      fflush(stdout);
      memset(k,0,sizeof(k));
      fgets(k,sizeof(k),stdin);
      if (k[0]=='y' || k[0]=='Y')
      {
        pcmd(0xACA80000| fb,4);
        continue;
      }
    }
    if (!strcmp(args[0],"resp"))
    {
     printf("Resetting queue pointers\n");
     Usimple0(CMD_RESETPTRS);
     continue;
    }
    if (!strcmp(args[0],"reprom"))
    {
      read_eeprom();
      continue;
    }
  }

done:
  Ufinish();
  return 0;
}
